@@ -5,7 +5,7 @@ import typing
 import dataclasses_json
 
 from ..json import DataJsonFormat, ValidatingDataJsonFormat, package_json_format
-from ..sql import SqlObject
+from ..sql import SqlId, SqlObject
 
 
 class JoinConsistency(enum.Enum):
@@ -13,23 +13,33 @@ class JoinConsistency(enum.Enum):
     IMMEDIATE = "immediate"
 
 
+class JoinDepMode(enum.Enum):
+    ASYNC = "async"
+    SYNC = "sync"
+
+
+class JoinSync(enum.Enum):
+    FULL = "full"
+    UPSERT = "upsert"
+
+
 @dataclasses_json.dataclass_json(
     letter_case=dataclasses_json.LetterCase.CAMEL,
     undefined=dataclasses_json.Undefined.EXCLUDE,
 )
 @dataclasses.dataclass
 class JoinTarget:
-    columns: typing.List[str]
     key: typing.List[str]
     name: str
     schema: typing.Optional[str] = None
+    columns: typing.Optional[typing.List[str]] = None
 
     @property
     def sql(self) -> SqlObject:
         return (
-            SqlObject(self.schema, self.name)
+            SqlObject(SqlId(self.schema), SqlId(self.name))
             if self.schema is not None
-            else SqlObject(self.name)
+            else SqlObject(SqlId(self.name))
         )
 
 
@@ -39,12 +49,14 @@ class JoinTarget:
 )
 @dataclasses.dataclass
 class JoinTable:
-    id: str
     name: str
     dep: typing.Optional[str] = None
+    dep_join: typing.Optional[str] = None
+    dep_mode: JoinDepMode = JoinDepMode.SYNC
     key: typing.Optional[typing.List[str]] = None
-    join: typing.Optional[str] = None
+    lock_id: typing.Optional[int] = None
     schema: typing.Optional[str] = None
+    target_key: typing.Optional[typing.List[str]] = None
 
     def __eq__(self, other):
         return self is other
@@ -55,9 +67,9 @@ class JoinTable:
     @property
     def sql(self) -> SqlObject:
         return (
-            SqlObject(self.schema, self.name)
+            SqlObject(SqlId(self.schema), SqlId(self.name))
             if self.schema is not None
-            else SqlObject(self.name)
+            else SqlObject(SqlId(self.name))
         )
 
 
@@ -73,39 +85,26 @@ class JoinHook:
     @property
     def sql(self) -> SqlObject:
         return (
-            SqlObject(self.schema, self.name)
+            SqlObject(SqlId(self.schema), SqlId(self.name))
             if self.schema is not None
-            else SqlObject(self.name)
+            else SqlObject(SqlId(self.name))
         )
 
 
-@dataclasses_json.dataclass_json(
-    letter_case=dataclasses_json.LetterCase.CAMEL,
-    undefined=dataclasses_json.Undefined.EXCLUDE,
-)
-@dataclasses.dataclass(frozen=True)
-class JoinHooks:
-    before: typing.Optional[JoinHook] = None
-
-
 @dataclasses_json.dataclass_json(
     letter_case=dataclasses_json.LetterCase.CAMEL,
     undefined=dataclasses_json.Undefined.EXCLUDE,
 )
 @dataclasses.dataclass
-class Join:
+class JoinConfig:
     id: str
-    query: str
-    tables: typing.List[JoinTable]
+    tables: typing.Dict[str, JoinTable]
     target: JoinTarget
     consistency: JoinConsistency = JoinConsistency.IMMEDIATE
-    hooks: JoinHooks = JoinHooks(before=None)
+    query: typing.Optional[str] = None
     schema: typing.Optional[str] = None
-
-    def sql_object(self, name):
-        return (
-            SqlObject(self.schema, name) if self.schema is not None else SqlObject(name)
-        )
+    setup: typing.Optional[JoinHook] = None
+    sync: JoinSync = JoinSync.FULL
 
 
 class JoinInvalid(Exception):
@@ -113,6 +112,17 @@ class JoinInvalid(Exception):
         super().__init__(message)
 
 
+def validate_join(join: JoinConfig):
+    if join.consistency == JoinConsistency.DEFERRED and join.query is None:
+        raise JoinInvalid("Deferrable mode is only used with query")
+
+    if join.target.columns is None and join.query is not None:
+        raise JoinInvalid("Query requires target column list")
+
+
 JOIN_JSON_FORMAT = package_json_format("denorm.formats", "join.json")
 
-JOIN_DATA_JSON_FORMAT = DataJsonFormat(JOIN_JSON_FORMAT, Join.schema())
+JOIN_DATA_JSON_FORMAT = ValidatingDataJsonFormat(
+    DataJsonFormat(JOIN_JSON_FORMAT, JoinConfig.schema()),
+    validate_join,
+)
